/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <fstream>
using std::ifstream;

#include <algorithm>
using std::sort;

#include "AVTree.h"
#include "Attribute.h"
#include "Item.h"
#include "Cover.h"
#include "FITree.h"

// Standard output function
ostream& operator<<(ostream& os, const AVTree& tree) {
	os << "Attributes:" << endl;
	for(int i = 0; i < tree.attributes.size(); i++) {
		os << *(tree.attributes[i]) << endl;
	}

	return os;
}

/**
 * Create an empty attribute-value tree.
 */
AVTree::AVTree() {
	transactions = 0;
}

/**
 * Create an attribute-value tree from a text file.
 * @param data_file The name of the text file.
 * @param column_separator The character separating the columns (typically '\t').
 * @param transaction_separator The character separating the transactions (typically '\n').
 */
AVTree::AVTree(string data_file, char column_separator, char transaction_separator) {
	transactions = 0;
	buildTree(data_file, column_separator, transaction_separator);
}

/**
 * Create an attribute-value tree from a text file.
 * @param data_file The name of the text file.
 * @param column_separator The character separating the columns (typically '\t').
 * @param transaction_separator The character separating the transactions (typically '\n').
 * @param transaction_data The data is in transaction format.
 */
AVTree::AVTree(string data_file, char column_separator, char transaction_separator, bool transaction_dataset) {
	if (transaction_dataset) {
		transactions = 0;
		buildTreeTransaction(data_file, column_separator, transaction_separator);
	} else {
		transactions = 0;
		buildTree(data_file, column_separator, transaction_separator);
	}
}

/**
 * The destructor.
 */
AVTree::~AVTree() {
	for (int i = 0; i < attributes.size(); i++) {
		delete attributes[i];
	}
}

/**
 * Build an attribute-value tree from a text file.
 * @param data_file The name of the text file.
 * @param column_separator The character separating the columns (typically '\t').
 * @param transaction_separator The character separating the transactions (typically '\n').
 */
void AVTree::buildTree(string file_name, char column_separator, char transaction_separator) {
	buildTree(file_name.c_str(), column_separator, transaction_separator);
}

/**
 * Build an attribute-value tree from a text file.
 * @param data_file The name of the text file.
 * @param column_separator The character separating the columns (typically '\t').
 * @param transaction_separator The character separating the transactions (typically '\n').
 */
void AVTree::buildTree(const char* file_name, char column_separator, char transaction_separator) {
	ifstream database(file_name);
	
	if (!database.is_open()) {
		std::cerr << endl << "Error opening file: " << file_name << endl;
		exit(1);
	}

	char c;

	string value;
	int column = 0;

	int BUFFER_SIZE = 1000000;
	char* buffer = new char[BUFFER_SIZE];
	
	database.getline(buffer, BUFFER_SIZE, transaction_separator);
	int chars_read = database.gcount();
	
	// Read the attributes (headers of columns)
	for (int i = 0; i < chars_read - 1; i++) {
		c = buffer[i];
		//std::cerr << "c:" << (int)c << std::endl;
		
		if (c == column_separator) {
			attributes.push_back(new Attribute(value, column));
			column++;
			value = "";
		} else {
			value.append(1, c);
		}
	}
	
	attributes.push_back(new Attribute(value, column));

	// Read the transactions
	value = "";
	column = 0;
	
	do {
		database.read(buffer, BUFFER_SIZE);
		chars_read = database.gcount();
		
		for (int i = 0; i < chars_read; i++) {
			c = buffer[i];
			
			if (c == column_separator) {
				//std::cerr << "Adding: v:'" << value << "' t:'" << transaction << "'" << endl;
				if (value != "") {
					attributes[column]->addValue(value, transactions);
				}
				column++;
				value = "";
			} else if (c == transaction_separator && column > 0) {
				//std::cerr << "Adding: v:'" << value << "' t:'" << transaction << "'" << endl;
				if (value != "") {
					attributes[column]->addValue(value, transactions);
				}
				column = 0;
				transactions++;
				value = "";
			} else {
				value.append(1, c);
			}
		}
	} while (!database.eof());
	
	database.close();
}

/**
 * Build an attribute-value tree from a text file containing transaction dataset.
 * @param data_file The name of the text file.
 * @param column_separator The character separating the columns (typically '\t').
 * @param transaction_separator The character separating the transactions (typically '\n').
 */
void AVTree::buildTreeTransaction(string file_name, char column_separator, char transaction_separator) {
	buildTreeTransaction(file_name.c_str(), column_separator, transaction_separator);
}

/**
 * Build an attribute-value tree from a text file containing transaction dataset.
 * @param data_file The name of the text file.
 * @param column_separator The character separating the columns (typically '\t').
 * @param transaction_separator The character separating the transactions (typically '\n').
 */
void AVTree::buildTreeTransaction(const char* file_name, char column_separator, char transaction_separator) {
	ifstream database(file_name);
	
	if (!database.is_open()) {
		std::cerr << std::endl << "Error opening file: " << file_name << std::endl;
		exit(1);
	}

	char c;

	string value;

	int BUFFER_SIZE = 1000000;
	char* buffer = new char[BUFFER_SIZE];
	
	int chars_read;
	
	attributes.push_back(new Attribute("Item", 0));
	
	// Read the transactions
	do {
		database.read(buffer, BUFFER_SIZE);
		chars_read = database.gcount();
		
		for (int i = 0; i < chars_read; i++) {
			c = buffer[i];
			
			if (c == column_separator) {
				//std::cerr << "Adding: v:'" << value << "' t:'" << transaction << "'" << endl;
				if (value != "") {
					attributes[0]->addValue(value, transactions);
				}
				value = "";
			} else if (c == transaction_separator) {
				//std::cerr << "Adding: v:'" << value << "' t:'" << transaction << "'" << endl;
				if (value != "") {
					attributes[0]->addValue(value, transactions);
				}
				transactions++;
				value = "";
			} else {
				value.append(1, c);
			}
		}
	} while (!database.eof());
	
	database.close();
}

/**
 * Get the number of transactions in this dataset
 * @return The number of transactions.
 */
int AVTree::getTransactions() {
	return transactions;
}

/**
 * Get the frequent items in this dataset.
 * @param The minimum support.
 * @return The root node of the frequent itemset tree containing the frequent items (1-itemsets).
 */
FITree* AVTree::frequentItems(int min_supp) {
	FITree* root = new FITree();
	root->support = transactions;
	root->height = 0;
	root->upper_bound = transactions;
	root->lower_bound = 0;

	root->tidhash = 0;
	for (int i = 0; i < transactions; i++) {
		root->tidhash = root->tidhash + i;
	}

	
	FITree* item;
	
	for (int i = 0; i < attributes.size(); i++) {
		for (map<string, Item*>::iterator iter = attributes[i]->items.begin(); iter != attributes[i]->items.end(); iter++) {
			if (iter->second->cover->size >= min_supp) {
				item = new FITree(iter->second, root, iter->second->cover, iter->second->cover->size);
				item->height = 1;
				item->upper_bound = transactions;
				item->lower_bound = 0;
				item->tidhash = iter->second->cover->tidhash();
				root->addChild(item);
			}
		}
	}
	
	// Sorting the initial 1-itemsets in ascending order of support will result in
	// more cases of support of a superset being equal to the support of the itemset
	sort((root->children).begin(), (root->children).end(), fi_tree_pointer_compare);
	
	return root;
}
