/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ConfidenceConstraint.h"
#include "FITree.h"
#include "AssociationRule.h"

/**
 * Create a confidence constraint with minimum confidence of 0
 */
ConfidenceConstraint::ConfidenceConstraint() {
	last_confidence = 0;
	min_confidence = 0;
}

/**
 * Create a confidence constraint with a minimum confidence.
 * @param d The minimum confidence.
 */
ConfidenceConstraint::ConfidenceConstraint(double d) {
	last_confidence = 0;
	min_confidence = d;
}

/**
 * Set the minimum confidence of this constraint.
 * @param d The minimum confidence.
 */
void ConfidenceConstraint::setMinConfidence(double d) {
	min_confidence = d;
}

/**
 * Returns the result of last evaluation.
 * @return The confidence of the pattern evaluated last.
 */
double ConfidenceConstraint::lastResult() {
	return last_confidence;
}

/**
 * Value an association rule.
 * @param arule The association rule.
 * @return The confidence of the association rule.
 */
double ConfidenceConstraint::value(AssociationRule* arule) {
	last_confidence = (double)arule->itemset->support / (double)arule->condition->support;
	return last_confidence;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the confidence of the rule is higher than the minimum
 * confidence of this constraint.
 */
bool ConfidenceConstraint::validate(AssociationRule* arule) {
	if (value(arule) >= min_confidence) {
		return true;
	}
	
	return false;
}

/**
 * Returns the minimum confidence.
 * @return The minimum confidence.
 */
double ConfidenceConstraint::constraintLimit() const {
	return min_confidence;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string ConfidenceConstraint::name() const {
	return "Confidence";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string ConfidenceConstraint::abbreviation() const {
	return "c";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string ConfidenceConstraint::description() const {
	return "Confidence is the support of the union of the consequent and condition of the rule divided by the support of the condition of the rule.";
}
