/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <iostream>
using std::pair;
using std::make_pair;

#include "Eclat.h"
#include "AVTree.h"
#include "ContextTree.h"
#include "FITree.h"
#include "Item.h"
#include "Cover.h"

/**
 * Create a new miner. Use diffsets and FITree compression.
 */
Eclat::Eclat() {
	diffsets = true;
	expand_all_itemsets = false;
}

/**
 * Create a new miner. Use of diffsets and FITree compression is defined.
 * @param d Usage of diffsets. True will set on the usage of diffsets
 * @param e Usage of FITree compression. True will set off the compression.
 */
Eclat::Eclat(bool d, bool e) {
	diffsets = d;
	expand_all_itemsets = e;
}

/**
 * The destructor.
 */
Eclat::~Eclat() {
}

/**
 * Set the usage of diffsets
 * @param d Usage of diffsets. True will set on the usage of diffsets
 */
void Eclat::setDiffsets(bool d) {
	diffsets = d;
}

/**
 * Set the usage of FITree compression
 * @param e Usage of FITree compression. True will set off the compression.
 */
void Eclat::setExpand(bool e) {
	expand_all_itemsets = e;
}

/**
 * Generate all frequent itemsets from attribute-value data with given support threshold
 * @param dataset The dataset.
 * @param min_support The support threshold.
 */
FITree* Eclat::generateFrequentSets(AVTree* dataset, int min_support) {
	FITree* frequent_items = dataset->frequentItems(min_support);

 	generateFrequentSets(frequent_items, min_support);

	return frequent_items;
}

/**
 * Generate all frequent itemsets from context data with given support threshold
 * @param dataset The dataset.
 * @param min_support The support threshold.
 */
FITree* Eclat::generateFrequentSets(ContextTree* dataset, int min_support) {
	FITree* frequent_items = dataset->frequentItems(min_support);

 	generateFrequentContextSets(dataset, frequent_items, min_support);

	return frequent_items;
}

/**
 * Generate all frequent closed itemsets from attribute-value data with given support threshold
 * @param dataset The dataset.
 * @param min_support The support threshold.
 */
FITree* Eclat::generateFrequentClosedSets(AVTree* dataset, int min_support) {
	FITree* frequent_items = dataset->frequentItems(min_support);
	expand_all_itemsets = false;

	for (int i = 0; i < frequent_items->root()->children.size(); i++) {
		closed_itemsets.insert(make_pair(frequent_items->root()->children[i]->tidhash,frequent_items->root()->children[i]));
	}
	closed_itemsets.insert(make_pair(frequent_items->root()->tidhash,frequent_items->root()));
 	generateFrequentClosedSets(frequent_items, min_support);

	return frequent_items;
}

/**
 * Generate all frequent closed itemsets from context data with given support threshold
 * @param dataset The dataset.
 * @param min_support The support threshold.
 */
FITree* Eclat::generateFrequentClosedContextSets(ContextTree* dataset, int min_support) {
	FITree* frequent_items = dataset->frequentItems(min_support);
	expand_all_itemsets = false;

	for (int i = 0; i < frequent_items->root()->children.size(); i++) {
		closed_itemsets.insert(make_pair(frequent_items->root()->children[i]->tidhash,frequent_items->root()->children[i]));
	}
	closed_itemsets.insert(make_pair(frequent_items->root()->tidhash,frequent_items->root()));
 	generateFrequentClosedContextSets(dataset, frequent_items, min_support);

	return frequent_items;
}

/**
 * Recursively generate all frequent itemsets from attribute-value data with given support threshold
 * @param itemset The itemset iterator.
 * @param min_support The support threshold.
 */
void Eclat::generateFrequentSets(FITree* itemset, int min_support) {
	int number_of_children = (itemset->children).size();
	
	FITree* fset;
	Cover* cov;
	int sup;
	// Delete the cover if the itemset is not the root or one of the singleton itemsets (singleton itemsets have pointers to the
	// true covers of the itemsets
	if (itemset->cover != NULL && itemset->parent->parent != NULL) {
		delete itemset->cover;
	}
	
	for (int i = 0; i < number_of_children; i++) {
		for (int j = i + 1; j < number_of_children; j++) {

			if (diffsets) {
				// dECLAT (eclat using diffsets)
				if (itemset->parent == NULL) {
					// Size 2 candidates
					cov = differenceCover(itemset->children[i]->cover, itemset->children[j]->cover);
				} else {
					cov = differenceCover(itemset->children[j]->cover, itemset->children[i]->cover);					
				}
				sup = (itemset->children[i])->support - cov->size;
			}	else { 
				// ECLAT
				cov = intersectionCover(itemset->children[i]->cover, itemset->children[j]->cover);
				sup = cov->size;
			}

			if (itemset->children[i]->support == sup && !expand_all_itemsets) {
				itemset->children[i]->addItem((itemset->children[j])->items[0]);
				delete cov;
			} else if (sup >= min_support) {
				fset = new FITree((itemset->children[j])->items[0], itemset->children[i], cov, sup);
				itemset->children[i]->addChild(fset);
			} else {
				delete cov;
			}
		}
		
		generateFrequentSets(itemset->children[i], min_support);
	}	
}

/**
 * Recursively generate all frequent itemsets from context data with given support threshold
 * @param dataset The context dataset
 * @param itemset The itemset iterator.
 * @param min_support The support threshold.
 */
void Eclat::generateFrequentContextSets(ContextTree* dataset ,FITree* itemset, int min_support) {
	int number_of_children = (itemset->children).size();
	
	FITree* fset;
	Cover* cov;
	int sup;
	// Delete the cover if the itemset is not the root or one of the singleton itemsets (singleton itemsets have pointers to the
	// true covers of the itemsets
	if (itemset->cover != NULL && itemset->parent->parent != NULL) {
		delete itemset->cover;
	}
	
	for (int i = 0; i < number_of_children; i++) {
		for (int j = i + 1; j < number_of_children; j++) {

			if (diffsets) {
				// dECLAT (eclat using diffsets)
				if (itemset->parent == NULL) {
					// Size 2 candidates
					cov = differenceCover(itemset->children[i]->cover, itemset->children[j]->cover);
				} else {
					cov = differenceCover(itemset->children[j]->cover, itemset->children[i]->cover);					
				}
				sup = (itemset->children[i])->support - dataset->contextSupport(cov);
			}	else { 
				// ECLAT
				cov = intersectionCover(itemset->children[i]->cover, itemset->children[j]->cover);
				sup = dataset->contextSupport(cov);
			}

			if (itemset->children[i]->support == sup && !expand_all_itemsets) {
				itemset->children[i]->addItem((itemset->children[j])->items[0]);
				delete cov;
			} else if (sup >= min_support) {
				fset = new FITree((itemset->children[j])->items[0], itemset->children[i], cov, sup);
				itemset->children[i]->addChild(fset);
			} else {
				delete cov;
			}
		}
		
		generateFrequentContextSets(dataset, itemset->children[i], min_support);
	}	
}

/**
 * Recursively generate all frequent closed itemsets from attribute-value data with given support threshold
 * @param itemset The itemset iterator.
 * @param min_support The support threshold.
 */
void Eclat::generateFrequentClosedSets(FITree* itemset, int min_support) {
	FITree* fset;
	Cover* cov;
	int sup;
	int tidhash;
	
	// Delete the cover if the itemset is not the root or one of the singleton itemsets (singleton itemsets have pointers to the
	// true covers of the itemsets
	if (itemset->cover != NULL && itemset->parent->parent != NULL) {
		delete itemset->cover;
	}
			
	for (int i = 0; i < (itemset->children).size(); i++) {
		for (int j = i + 1; j < (itemset->children).size(); j++) {

			if (diffsets) {
				// dECLAT (eclat using diffsets)
				if (itemset->parent == NULL) {
					// Size 2 candidates
					cov = differenceCover(itemset->children[i]->cover, itemset->children[j]->cover);
				} else {
					cov = differenceCover(itemset->children[j]->cover, itemset->children[i]->cover);					
				}
				sup = (itemset->children[i])->support - cov->size;
				tidhash = (itemset->children[i])->tidhash - cov->tidhash();
			}	else { 
				// ECLAT
				cov = intersectionCover(itemset->children[i]->cover, itemset->children[j]->cover);
				tidhash = cov->tidhash();
				sup = cov->size;
			}


			if (sup >= min_support) {
				if (itemset->children[i]->support == sup) {
					// Merge the newly found itemset with the old one with same support
					itemset->children[i]->addItem((itemset->children[j])->items[0]);
					delete cov;
				} else {
					// Create a new child
					fset = new FITree(itemset->children[j]->items[0], itemset->children[i], cov, sup);
					fset->tidhash = tidhash;

					// Do a hashcode check to see whether there are already supersets with equal support
					if (!hasClosedSuperset(fset)) {
						itemset->children[i]->addChild(fset);
						closed_itemsets.insert(make_pair(tidhash,fset));
					} else {
						delete cov;
						delete fset;
					}
				}

				// When mining closed itemsets delete the second iterator if it has the same support as
				// the newly created/merged itemset
				if(itemset->children[j]->support == sup) {
					if (itemset->children[j]->cover != NULL && itemset->children[j]->parent->parent != NULL) {
						delete itemset->children[j]->cover;
					}
					
					FITree* child = itemset->removeChild(j);
					removeClosedSet(child);
					delete child;
					j--;
				}

			} else {
				delete cov;
			}
		}
		generateFrequentClosedSets(itemset->children[i], min_support);
	}	
}

/**
 * Recursively generate all frequent closed itemsets from context data with given support threshold
 * @param dataset The context dataset.
 * @param itemset The itemset iterator.
 * @param min_support The support threshold.
 */
void Eclat::generateFrequentClosedContextSets(ContextTree* dataset, FITree* itemset, int min_support) {
	FITree* fset;
	Cover* cov;
	int sup;
	int tidhash;
	
	// Delete the cover if the itemset is not the root or one of the singleton itemsets (singleton itemsets have pointers to the
	// true covers of the itemsets
	if (itemset->cover != NULL && itemset->parent->parent != NULL) {
		delete itemset->cover;
	}
			
	for (int i = 0; i < (itemset->children).size(); i++) {
		for (int j = i + 1; j < (itemset->children).size(); j++) {

			if (diffsets) {
				// dECLAT (eclat using diffsets)
				if (itemset->parent == NULL) {
					// Size 2 candidates
					cov = differenceCover(itemset->children[i]->cover, itemset->children[j]->cover);
				} else {
					cov = differenceCover(itemset->children[j]->cover, itemset->children[i]->cover);					
				}
				sup = (itemset->children[i])->support - dataset->contextSupport(cov);
				tidhash = (itemset->children[i])->tidhash - cov->tidhash();
			}	else { 
				// ECLAT
				cov = intersectionCover(itemset->children[i]->cover, itemset->children[j]->cover);
				tidhash = cov->tidhash();
				sup = dataset->contextSupport(cov);
			}


			if (sup >= min_support) {
				if (itemset->children[i]->support == sup) {
					// Merge the newly found itemset with the old one with same support
					itemset->children[i]->addItem((itemset->children[j])->items[0]);
					delete cov;
				} else {
					// Create a new child
					fset = new FITree(itemset->children[j]->items[0], itemset->children[i], cov, sup);
					fset->tidhash = tidhash;

					// Do a hashcode check to see whether there are already supersets with equal support
					if (!hasClosedSuperset(fset)) {
						itemset->children[i]->addChild(fset);
						closed_itemsets.insert(make_pair(tidhash,fset));
					} else {
						delete cov;
						delete fset;
					}
				}

				// When mining closed itemsets delete the second iterator if it has the same support as
				// the newly created/merged itemset
				if(itemset->children[j]->support == sup) {
					if (itemset->children[j]->cover != NULL && itemset->children[j]->parent->parent != NULL) {
						delete itemset->children[j]->cover;
					}
					
					FITree* child = itemset->removeChild(j);
					removeClosedSet(child);
					delete child;
					j--;
				}

			} else {
				delete cov;
			}
		}
		generateFrequentClosedContextSets(dataset, itemset->children[i], min_support);
	}	
}

/**
 * Checks if an itemset has a closed superset with same support.
 * @param itemset The itemset to be checked for closed supersets.
 * @return True if the itemset has a closed superset with same support.
 */
bool Eclat::hasClosedSuperset(FITree* itemset) {
	pair<multimap<int, FITree*>::iterator, multimap<int, FITree*>::iterator> super_set_candidates = closed_itemsets.equal_range(itemset->tidhash);

	for (multimap<int, FITree*>::iterator i = super_set_candidates.first; i != super_set_candidates.second; i++) {
		if (i->second->isSuperset(itemset) && i->second->support == itemset->support) {
			return true;
		}
	}
	
	return false;
}

/**
 * Remove an itemset from the closed itemset mapping.
 * @param itemset the itemset to be removed
 */
void Eclat::removeClosedSet(FITree* itemset) {
	pair<multimap<int, FITree*>::iterator, multimap<int, FITree*>::iterator> super_set_candidates = closed_itemsets.equal_range(itemset->tidhash);

	for (multimap<int, FITree*>::iterator i = super_set_candidates.first; i != super_set_candidates.second; i++) {
		if (i->second == itemset) {
			closed_itemsets.erase(i);
			break;
		}
	}
}
