/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef FIRM_FITREE
#define FIRM_FITREE
#define _STORE_SUBSETS

#include <iostream>
using std::ostream;

#include <vector>
using std::vector;

#include <map>
using std::map;

#include "AssociationRule.h"

class Item;
class Attribute;
class Cover;

/**
 * A data structure for holding itemsets. FITree is a prefix tree in which itemsets are
 * stored so that an itemset contains all items from the node representing it to the root
 * of the tree.
 * @author Juho Muhonen
 * @modified 10.10.2005 commented .cpp
 * @modified 7.10.2005 commented .h
 */
class FITree {
private:
#ifdef _STORE_SUBSETS
	/**
	 * Subsets of this itemset
	 */
	vector<FITree*>* _subsets;
#endif

public:
	/**
	 * The parent of this itemset.
	 */
	FITree* parent;
	
	/**
	 * A holder for cover calculations. This variable is used for example for
	 * holding diffsets when mining frequent itemsets with eclat-algorithm
	 */
	Cover* cover;
	
	/**
	 * The support of the itemset
	 */
	int support;
	
	/**
	 * The height is the steps from this itemset to the root node. Root node has a height of 0.
	 */
	int height;
	
	/**
	 * A variable for storing a hash value counted from the tids in the cover of this itemset.
	 */
	int tidhash;
	
	/**
	 * A variable for storing the upper bound of support in NDI-algorithm
	 */
	int upper_bound;

	/**
	 * A variable for storing the lower bound of support in NDI-algorithm
	 */
	int lower_bound;
	
	/**
	 * Items in this node.
	 */
	vector<Item*> items;
	
	/**
	 * Children of this node.
	 */
	vector<FITree*> children;
	
	/**
	 * A container for storing association rules that have this itemset as its condition for NDAR-algorithm
	 */
	map<FITree*, AssociationRule*> condition_rules;

	/**
	 * A container for storing association rules that have this itemset as its consequent for NDAR-algorithm
	 */
	map<FITree*, AssociationRule*> consequent_rules;	
public:
	/**
	 * Total number of created itemsets. For debugging purposes.
	 */
	static int CREATIONS;

	/**
	 * Total number of deleted itemsets. For debugging purposes.
	 */
	static int DELETIONS;

	/**
	 * Creates a new root node.
	 */
	FITree();
	
	/**
	 * Creates a new root node with an item stored.
	 */
	FITree(Item*);
	
	/**
	 * Creates a new node with parent and an item stored.
	 */
	FITree(Item*, FITree*);
	
	/**
	 * Creates a new node with parent and cover and an item stored.
	 */
	FITree(Item*, FITree*, Cover*);
	
	/**
	 * Creates a new node with parent, cover and support and an item stored.
	 */
	FITree(Item*, FITree*, Cover*, int);
	
	/**
	 * Creates a copy of a tree (node and it's children recursively)
	 */
	FITree(const FITree&);
	
	/**
	 * The destructor.
	 */
	~FITree();

	/**
	 * Tries to find an itemset using this node as the root.
	 */
	FITree* findItemset(vector<Item*>*);
	
	/**
	 * Tries to find the child containing a given item (or returns this node, if this node contains the item)
	 */
	FITree* findChild(Item*);
	
	/**
	 * Tries to find an item from this node.
	 */
	FITree* findItem(Item*);
	
	/** 
	 * Returns the root of the tree this node belongs to.
	 */
	FITree* root();

	/**
	 * Adds a child to this node.
	 */
	void addChild(FITree*);
	
	/**
	 * Removes the last child
	 */
	FITree* removeChild();
	
	/**
	 * Removes a child in given location.
	 */
	FITree* removeChild(int);
	
	/**
	 * Adds an item to this node.
	 */
	void addItem(Item*);
	
	/**
	 * Clears the items of this node.
	 */
	void clearItems();
	
	/**
	 * Clears the children of this node
	 */
	void clearChildren();
	
	/**
	 * Actually calculates the support of this itemset using the covers of the stored items.
	 */
	void countSupport();
	
	/**
	 * Returns the cover of this itemset. That is the intersection of the item covers.
	 */
	Cover* getCover();
		
	/**
	 * Tells whether this itemset is a superset of a given itemset.
	 */
	bool isSuperset(FITree*);
	
	/**
	 * Checks if all the subsets of this itemset are frequent.
	 */
	bool subsetsFrequent();
	
	/**
	 * Checks if this itemset has an item with a given attribute.
	 */
	bool hasAttribute(Attribute*) const;

	/**
	 * Checks if this itemset has an item.
	 */
	bool hasItem(Item*);
	
	/**
	 * Counts the number of frequent itemsets from this node onward.
	 */
	int numberOfFrequentSets() const;
	
	/**
	 * Counts the number of nodes from this node onward.
	 */
	int numberOfNodes() const;
	
	/**
	 * Returns the size of this itemset
	 */
	int size();
	
	/**
	 * Returns the number of items in this node.
	 */
	int itemCount();
	
	/**
	 * Compares two nodes.
	 */
	int operator<(const FITree&) const;
	
	/**
	 * Outputs this the tree from this node onward to an output stream
	 */
	void outputTree(ostream&);
	
	/**
	 * Outputs all itemsets from this node onward to an output stream
	 */
	void outputItemsets(ostream&);
	
	void outputItems(ostream&);
	void outputItems(ostream&, FITree*);

	/**
	 * Expands this FITree if "the same support optimization2 has been used
	 */
	void expand();
	
	/**
	 * Returns the subsets of this itemset.
	 */
	vector<FITree*>* subsets();
	
	/**
	 * Returns all association rules that can be generated from this itemset.
	 */
	vector<AssociationRule*>* associationRules();
	
	/**
	 * A helper function for subsets()
	 */
	void subsets(int, vector<Item*>*, vector<FITree*>*);

	/**
	 * Finds an union itemset.
	 */
	FITree* unionItemset(FITree*);
	
	/**
	 * Standard out
	 */
	friend ostream& operator<<(ostream&, const FITree&);
	
private:
	void outputItemsets(vector<Item*>*, ostream&, int);
	void outputItemsets(vector<Item*>*, ostream&);
};

int fi_tree_pointer_compare(const FITree*, const FITree*);
int fi_tree_pointer_support_compare(const FITree*, const FITree*);

ostream& operator<<(ostream&, const FITree&);
#endif
