/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ParameterParser.h"

/**
 * Prints a parameter parser to a ostream.
 */
ostream& operator<<(ostream& os, const ParameterParser& parser) {
	map<string, string>::const_iterator iter;
	for(iter = parser.parameters.begin(); iter != parser.parameters.end(); iter++) {
		os << iter->first << ": " << iter->second << endl;
	}
}

/**
 * A constant value given for flag parameters
 */
const string ParameterParser::FLAG_PARAMETER = "<FLAGGED>";

/**
 * Creates an empty parameter parser.
 */
ParameterParser::ParameterParser() {
}

/**
 * Creates a parameter parser from main(int argc, char* argv[]) type of input.
 * @param argc The number of parameters
 * @param argv An array of char arrays of the parameters.
 */
ParameterParser::ParameterParser(int argc, char* argv[]) {
  parseParameters(argc, argv);
}

/**
 * The destructor.
 */
ParameterParser::~ParameterParser() {
}

/**
 * Adds a parameter to this parser.
 * @param key The parameter name.
 * @param value The parameter value.
 */
void ParameterParser::addParameter(string key, string value) {
	parameters[key] = value;
}

/**
 * Parses parameters from main(int argc, char* argv[]) type of input.
 * @param argc The number of parameters
 * @param argv An array of char arrays of the parameters.
 */
void ParameterParser::parseParameters(int argc, char* argv[]) {
	bool flag = false;
	string args;
	string argfirst;
	string param;
	
	for (int i = 1; i < argc; i++) {
		args = argv[i];
		argfirst = args.substr(0,1);
	
		if (flag) {
			if (argfirst == "-") {
				parameters[param] = ParameterParser::FLAG_PARAMETER;
				flag = false;				
			} else {
				parameters[param] = args;
				flag = false;
			}
		}

		if (argfirst == "-") {
			param = args.substr(1,args.length());
			flag = true;
		} 
	}
	
	if (flag) {
		if (argfirst == "-") {
			parameters[param] = ParameterParser::FLAG_PARAMETER;
			flag = false;				
		} else {
			parameters[param] = args;
			flag = false;
		}
	}

}

/**
 * Checks for a presence of a parameter.
 * @param key The parameter name.
 * @return True if this parser has the parameter. False otherwise.
 */
bool ParameterParser::hasParameter(string key) {
	if (parameters.find(key) == parameters.end()) {
		return false;
	}
	
	return true;
}

/**
 * Checks if a parameter is only a flag(it has no actual value) parameter.
 * @param key The parameter name.
 * @return True if the given parameter is a flag. False if there is no such parameter, or the parameter has actual value.
 */
bool ParameterParser::isFlagParameter(string key) {
	if (parameters.find(key) == parameters.end()) {
		return false;
	}
	
	if (parameters[key] == ParameterParser::FLAG_PARAMETER) {
		return true;
	}
	
	return false;
}


/**
 * Returns a parameter value
 * @param key The parameter name.
 * @return The corresponding parameter value. NULL if no such parameter has been defined.
 */
string ParameterParser::getParameter(string key) {
	if (parameters.find(key) == parameters.end()) {
		return NULL;
	}
	
	return parameters[key];
}
