/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef FIRM_RULECONSTRAINT
#define FIRM_RULECONSTRAINT

#include <string>
using std::string;

class AssociationRule;

/**
 * A Class that is the superclass of all rule constraints for generating customized output.
 * @author Juho Muhonen
 * @modified 31.10.2005 precision()
 * @modified 6.10.2005 commented all
 */
class RuleConstraint {
public:
	/**
	 * Return the value of the constraint for a given association rule.
	 */
	virtual double value(AssociationRule*) {return 0;};
	
	/**
	 * Validate a given association rule. Validate should use function value() when deciding
	 * on the validation.
	 */
	virtual bool validate(AssociationRule*) {return true;};

	/**
	 * Return the last result of a call to value or validate.
	 */
	virtual double lastResult() const {return 0;}

	/**
	 * Return the constraint limit of the constraint (the minimum or maximum value fo which a given set is valid).
	 */
	virtual double constraintLimit() const {return 0;}

	/**
	 * Return the highest possible value of this constraint. This can be used when coloring a value of an itemset
	 * according to its relative location in the parameter space. For example an itemset with frequency close to
	 * 1 (maximum frequency is 1) can now be colored differently from an itemset with lower frequency.
	 */
	virtual double upperBound() const {return 1;}

	/**
	 * Return the lowest possible value of this constraint. This can be used when coloring a value of an itemset
	 * according to its relative location in the parameter space. For example an itemset with frequency close to
	 * 0 (minimum frequency is 0) can now be colored differently from an itemset with higher frequency.
	 */
	virtual double lowerBound() const {return -1;}

	/**
	 * Return the "zero point" of this constraint. This can be used when coloring a value of an itemset
	 * according to its relative location in the parameter space. For example an itemset with lift close to
	 * 1 ("normal" lift is 1) can now be colored differently from an itemset with higher or lower lift.
	 */
	virtual double zeroLimit() const {return 0;}

	/**
	 * Return the prefered precision of the constraint.
	 */
	virtual int precision() const {return 3;}

	/**
	 * Should return true, if it is better to sort the results in increasing order when using this constraint
	 * as criterion.
	 */
	virtual bool increasing() const {return true;}

	/**
	 * Return the name of the constraint.
	 */
	virtual string name() const {return "";}

	/**
	 * Return the abbreviation of the constraint.
	 */
	virtual string abbreviation() const {return "";}

	/**
	 * Return the description of the constraint.
	 */
	virtual string description() const {return "";}
};
#endif
