/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
     
#include "SupportConstraint.h"
#include "FITree.h"
#include "AssociationRule.h"

/**
 * Create a support constraint with minimum support of 0
 */
SupportConstraint::SupportConstraint() {
	last_support = 0;
	min_support = 0;
}

/**
 * Create a support constraint with a minimum support.
 * @param d The minimum support.
 */
SupportConstraint::SupportConstraint(int d) {
	last_support = 0;
	min_support = d;
}

/**
 * Set the minimum support of this constraint.
 * @param d The minimum support.
 */
void SupportConstraint::setMinSupport(int d) {
	min_support = d;
}

/**
 * Returns the result of last evaluation.
 * @return The support of the pattern evaluated last.
 */
double SupportConstraint::lastResult() {
	return last_support;
}

/**
 * Value an itemset.
 * @param fset The itemset.
 * @return The support of the itemset.
 */
double SupportConstraint::value(FITree* fset) {
	last_support = fset->support;
	return (double)last_support;
}

/**
 * Value an association rule.
 * @param arule The association rule.
 * @return The support of the association rule.
 */
double SupportConstraint::value(AssociationRule* arule) {
	last_support = arule->itemset->support;
	return (double)last_support;
}

/**
 * Validate an itemset.
 * @param fset The itemset.
 * @return True if the support of the itemset is higher than the minimum
 * support of this constraint.
 */
bool SupportConstraint::validate(FITree* fset) {
	if (value(fset) >= min_support) {
		return true;
	}
	
	return false;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the support of the rule is higher than the minimum
 * support of this constraint.
 */
bool SupportConstraint::validate(AssociationRule* arule) {
	if (value(arule) >= min_support) {
		return true;
	}
	
	return false;
}

/**
 * Returns 0.
 * @return 0
 */
int SupportConstraint::precision() const {
	return 0;
}

/**
 * Returns the minimum support.
 * @return The minimum support.
 */
double SupportConstraint::constraintLimit() const {
	return min_support;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string SupportConstraint::name() const {
	return "Support";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string SupportConstraint::abbreviation() const {
	return "s";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string SupportConstraint::description() const {
	return "Absolute frequency is the support of the pattern.";
}
